/*
	bt.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2006 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"

struct p_node* insert(struct p_node *n,int key, double count,int* size)
{
	int i;
	if(n == 0){
		n = malloc(sizeof(struct p_node));
		assert(n != 0);
		n->left = 0;
		n->right = 0;
		n->key = key;
		n->count = count;
		*size = *size +1 ;
	}else{
		i = n->key - key;
		if(i < 0){
			n->left = insert(n->left,key,count,size);
		}else if(i > 0){
			n->right = insert(n->right,key,count,size);
		}else if(i == 0){
			n->count += count;
			return n;
		}
	}
	return n;
}

void get_ann_score(struct p_node* n,double* out,int max_ann_no)
{
	int top;
	if(n->left){
		get_ann_score(n->left,out,max_ann_no);
	}
	if(n->key == -1){
		out[1] +=  n->count;
	}else{
		top  = pop(n->key);
		if(top){
			if(top <= max_ann_no){
				out[0] += n->count;
			}else{
				out[1] +=  n->count;
			}
		}
	}
	if(n->right){
		get_ann_score(n->right,out,max_ann_no);
	}
}


void print_tree(struct p_node* n,char** databases,double total,FILE *fout)
{
	int i,j,c,top;
	if(n->left){
		print_tree(n->left,databases,total,fout);
	}
	j = 1;
	c = 0;
	if(n->key == -1){
		fprintf(fout,"multi	");
	}else{
		top  = pop(n->key);
		if(top){
			for(i = 0; i < 32;i++){
				j = 1 << i;
				if(n->key & j){
					c++;
					if(c == top){
						fprintf(fout,"%s",shorten_name(databases[i]));
						break;
					}else{
						fprintf(fout,"%s,",shorten_name(databases[i]));
					}
				}
			}
		}else{
			fprintf(fout,"unannotated	");
		}
	}
	fprintf(fout,"	%0.2f	%0.2f\n",n->count / total *100.0,n->count);
	if(n->right){
		print_tree(n->right,databases,total,fout);
	}
}

int collect_pie_data(struct p_node* n,char** databases,char** colors,double* cat_expression,char** cat_name,char** cat_color,int cur,double total_expression)
{
	int i,j,c,top;
	char tmp_name[500];
	double tmp_exp = 0.0;
	tmp_name[0] = 0;
	if(n->left){
		cur = collect_pie_data(n->left,databases,colors,cat_expression,cat_name,cat_color,cur,total_expression);
	}
	j = 1;
	c = 0;
	cat_color[cur] = 0;
	if(n->key == -1){
		cat_color[cur] = "aquamarine";
		//cat_name[cur] = malloc(sizeof(char)*6);
		tmp_name[0] = 'a';
		tmp_name[1] = 'm';
		tmp_name[2] = 'b';
		tmp_name[3] = 'i';
		tmp_name[4] = 'g';
		tmp_name[5] = 'u';
		tmp_name[6] = 'o';
		tmp_name[7] = 'u';
		tmp_name[8] = 's';
		tmp_name[9] =  0;
	}else{
		top  = pop(n->key);
		if(colors){
			if(top == 1){
				for(i = 0; i < 32;i++){
					j = 1 << i;
					if(n->key & j){
						cat_color[cur] = colors[i];
					}
				}
			}else{
				cat_color[cur] = "white";
			}
		}else{
			cat_color[cur] = "white";
		}
		if(top){
			for(i = 0; i < 32;i++){
				j = 1 << i;
				if(n->key & j){
					c++;
					//fprintf(stderr,"Color = i + 1 - c + 1 ::: %d + 1 - %d + 1\n",i,c);
					//cat_color[cur] = i+1 - c  + 1;
					
					
					if(c == top){
						sprintf(tmp_name+strlen(tmp_name),"%s",shorten_name(databases[i]));
						break;
					}else{
						sprintf(tmp_name+strlen(tmp_name),"%s,",shorten_name(databases[i]));
					}
				}
			}
		}else{
			cat_color[cur] = "darkgoldenrod1";
			tmp_name[0] = 'u';
			tmp_name[1] = 'n';
			tmp_name[2] = 'a';
			tmp_name[3] = 'n';
			tmp_name[4] = 'n';
			tmp_name[5] = 'o';
			tmp_name[6] = 't';
			tmp_name[7] = 'a';
			tmp_name[8] = 't';
			tmp_name[9] = 'e';
			tmp_name[10] = 'd';
			tmp_name[11] = 0;
		}
	}
	tmp_exp = n->count/total_expression*100;
	if(tmp_exp >= 1.0){
		//fprintf(stderr,"Valid: %s	%f\n",tmp_name,tmp_exp);
		cat_name[cur] = malloc(sizeof(char)*strlen(tmp_name));
		for(i	= 0;i < strlen(tmp_name);i++){
			cat_name[cur][i] = tmp_name[i];
		}
		cat_name[cur][i] = 0;
		cat_expression[cur] = tmp_exp;
		cur++;
	}else{
		cat_expression[0] += tmp_exp;	
	}
	if(n->right){
		cur = collect_pie_data(n->right,databases,colors,cat_expression,cat_name,cat_color,cur,total_expression);
	}
	return cur;
}


void free_p_tree(struct p_node* n)
{
	if(n->left){
		free_p_tree(n->left);
	}
	if(n->right){
		free_p_tree(n->right);
	}
	free(n);
}


