/*
	output.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2007 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"
#include <ctype.h>
#include "sarray.h"

void write_bin_aln(struct fasta* p,const char* file)
{
	int fd;
	int i = 0;
	int test = 0;
	int mer_hash = 0;
	int version = SUFFIX_VERSION;
	int* len_array = malloc(sizeof(int)*p->numseq);
	assert(len_array !=0);
	for(i = 0; i < p->numseq;i++){
		len_array[i] = strlen(p->sn[i]);
	}
	if((fd = open(file,O_WRONLY)) == -1){
		creat(file,444);
		//fprintf(stderr,"Creating %s\n",file);
	}
	if((fd = open(file,O_WRONLY)) == -1){
		fprintf(stderr,"cannot open %s\n",file);
		exit(-1);
	}
	if(p->mer_hash){
		mer_hash = 1;
	}else{
		mer_hash = 0;
	}
	
	test = write(fd,&version,sizeof(int));
	assert(test >= 0);
	
	
	test = write(fd,&mer_hash,sizeof(int));
	assert(test >= 0);
	
	test = write(fd,&p->string_len,sizeof(int));
	assert(test >= 0);
	test = write(fd,&p->numseq,sizeof(int));
	assert(test >= 0);
	test = write(fd,&p->max_len,sizeof(int));
	assert(test >= 0);
	test = write(fd,len_array,sizeof(int)*p->numseq);
	assert(test >= 0);
	for(i = 0; i < p->numseq;i++){
		test = write(fd,p->sn[i],sizeof(char)*len_array[i]);
		assert(test >= 0);
	}
	
	test = write(fd,p->s_index,sizeof(int)*(p->numseq+1));
	assert(test >= 0);

	//test = write(fd,p->seq,sizeof(char*)*p->numseq);
	//assert(test >= 0);
 	test = write(fd,p->string,sizeof(char)*p->string_len+17);
	assert(test >= 0);
	test = write(fd,p->suffix,sizeof(int)*(p->string_len+1));
	assert(test >= 0);
	
	if(p->mer_hash){
		test = write(fd,p->mer_hash,sizeof(int)*134217728);
		assert(test >= 0);
	}
	
	close(fd);
	free(len_array);
}

void create_suffix(struct parameters* param)
{
	struct fasta* target = 0;
	char *time_s = 0;
	int db;
	int i;
   	
	for(db = 0;db < param->infiles;db++){
		if(!param->quiet_flag){
			fprintf(stderr,"Processing: %s\n",param->infile[db]);
			time_s = get_time(time_s);
		}
		if(param->solid){
			if(!param->quiet_flag){
				fprintf(stderr,"%s	Building SOLID suffix array...\n",time_s);
			}
			target = get_fasta(target,param->infile[db],param->solid);
			target->suffix = malloc((target->string_len+1)*sizeof(int));
			assert(target->suffix != 0);
			bsarray((unsigned char*)target->string, target->suffix, target->string_len);
			target->string = realloc(target->string,sizeof(char) * (target->string_len + 20));
			for (i = target->string_len+1;i < target->string_len+19;i++){
				target->string[i] = 'X';
			}
			target->string[i] = 0;
			target->string[target->string_len] = '!';
			
			if(target->string_len > 30000000){
				time_s = get_time(time_s);
				if(!param->quiet_flag){
				fprintf(stderr,"%s	Building hash...\n",time_s);
				}
				target->mer_hash = malloc(sizeof(int)* 134217728);
				assert(target->mer_hash != 0);
				for(i = 0;i < 134217728;i++){
					target->mer_hash[i] = 0;
				}
				
				//target->boost = malloc(sizeof(int) * 1048576*2);
				//for(i = 0;i < 1048576*2;i++){
				//	target->boost[i] = INT_MAX;
				//}
				
				
				mer_hash(target,16);
			}else{
				target->mer_hash = 0;
			}
			write_bin_aln(target,param->solid_suffixfile[db]);
		}else{
			if(!param->quiet_flag){
				fprintf(stderr,"%s	Building suffix array...\n",time_s);
			}
			target = get_fasta(target,param->infile[db],param->solid);
			target->suffix = malloc((target->string_len+1)*sizeof(int));
			assert(target->suffix != 0);
			bsarray((unsigned char*)target->string, target->suffix, target->string_len);
			target->string = realloc(target->string,sizeof(char) * (target->string_len + 20));
			for (i = target->string_len+1;i < target->string_len+19;i++){
				target->string[i] = 'X';
			}
			target->string[i] = 0;
			target->string[target->string_len] = '!';
			
			if(target->string_len > 30000000){
				time_s = get_time(time_s);
				if(!param->quiet_flag){
					fprintf(stderr,"%s	Building hash...\n",time_s);
				}
				target->mer_hash = malloc(sizeof(int)* 134217728);
				assert(target->mer_hash != 0);
				for(i = 0;i < 134217728;i++){
					target->mer_hash[i] = 0;
				}
				/*target->boost = malloc(sizeof(int) * 1048576*2);
				for(i = 0;i < 1048576*2;i++){
					target->boost[i] = INT_MAX;
				}*/
				
				
				mer_hash(target,16);
			}else{
				target->mer_hash = 0;
			}
			write_bin_aln(target,param->suffixfile[db]);
		}
		if(!param->quiet_flag){
			time_s = get_time(time_s);
			fprintf(stderr,"%s	Done \n",time_s);
		}
		free_aln(target);
	}
}

void print_stats(struct query_table* qt,struct parameters* param)
{
	int i,j;
	FILE *fout = NULL;
	
	char* outfile = param->stats;
	char* infile = param->fasta_file;
	
	char *names[] = { "Total","Exact","M","INDEL","MM","MMM"};
	
	if(outfile){
		if(!strcmp(outfile,"STDOUT")){
			fout = stdout; 
		}else{
			if ((fout = fopen(outfile, "w")) == NULL){
				fprintf(stderr,"can't open output\n");
				exit(-1);
			}
		}
	}else{
		fout = stdout; 
	}
	
	if(!infile){
		fprintf(fout,"Infile:STDIN\n");
	}else{
		fprintf(fout,"Infile:%s\n",infile);
	}

	fprintf(fout,"%1.0f total (%d unique)\n",qt->total_expression,qt->numseq);
	
	
	if(!param->upper_limit){
	
		fprintf(fout,"	Mapped	0	1	1-10	11-100	>100		Mapped	0	1	1-10	11-100	>100\n");
		for (i = 0; i < 6;i++){
			if(i ==2 && param->mismatches == 0){
			}else if(i == 1 && !(param->mode & 1)){
			}else if(i == 0 && param->mode == 1){
			}else if(i == 4 && param->mismatches < 2){
			}else if(i == 5 && param->mismatches < 3){
			}else if(i == 3 && !(param->mode & 4)){
			}else{
		
			fprintf(fout,"%s	",names[i]);
			fprintf(fout,"%0.2f%%	",(qt->stats[i][1]+qt->stats[i][2]+qt->stats[i][3]+qt->stats[i][4])/qt->total_expression*100);
			for(j = 0; j < 5;j++){
				fprintf(fout,"%0.2f%%	",qt->stats[i][j]/qt->total_expression*100);
			}
		
			fprintf(fout,"%s	",names[i]);
			fprintf(fout,"%0.2f	",qt->stats[i][1]+qt->stats[i][2]+qt->stats[i][3]+qt->stats[i][4]);
			for(j = 0; j < 5;j++){
				fprintf(fout,"%0.2f	",qt->stats[i][j]);
			}
			fprintf(fout,"\n");
			}
		}
	}else if(param->upper_limit == 1){
		fprintf(fout,"	Mapped	0	1	>1		Mapped	0	1	>1\n");
		/*for (i = 0; i < 6;i++){
			if(i ==2 && param->mismatches == 0){
			}else if(i == 1 && !(param->mode & 1)){
			}else if(i == 0 && param->mode == 1){
			}else if(i == 4 && param->mismatches < 2){
			}else if(i == 5 && param->mismatches < 3){
			}else if(i == 3 && !(param->mode & 4)){
			}else{*/
		i = 0;
		
				fprintf(fout,"%s	",names[i]);
				fprintf(fout,"%0.2f%%	",(qt->stats[i][1]+qt->stats[i][2])/qt->total_expression*100);
				for(j = 0; j < 3;j++){
					fprintf(fout,"%0.2f%%	",qt->stats[i][j]/qt->total_expression*100);
				}
		
				fprintf(fout,"%s	",names[i]);
				fprintf(fout,"%0.2f	",qt->stats[i][1]+qt->stats[i][2]);
				for(j = 0; j < 3;j++){
					fprintf(fout,"%0.2f	",qt->stats[i][j]);
				}
				fprintf(fout,"\n");
		//	}
		//}
	}else{
		fprintf(fout,"	Mapped	0	1	%d	>%d		Mapped	0	1	%d	>%d\n",param->upper_limit,param->upper_limit,param->upper_limit,param->upper_limit);
	/*for (i = 0; i < 6;i++){
			if(i ==2 && param->mismatches == 0){
			}else if(i == 1 && !(param->mode & 1)){
			}else if(i == 0 && param->mode == 1){
			}else if(i == 4 && param->mismatches < 2){
			}else if(i == 5 && param->mismatches < 3){
			}else if(i == 3 && !(param->mode & 4)){
			}else{*/
		i = 0;
		fprintf(fout,"%s	",names[i]);
		fprintf(fout,"%0.2f%%	",(qt->stats[i][1]+qt->stats[i][2]+qt->stats[i][3])/qt->total_expression*100);
		for(j = 0; j < 4;j++){
			fprintf(fout,"%0.2f%%	",qt->stats[i][j]/qt->total_expression*100);
		}
		
		fprintf(fout,"%s	",names[i]);
		fprintf(fout,"%0.2f	",qt->stats[i][1]+qt->stats[i][2]+qt->stats[i][3]);
		for(j = 0; j < 4;j++){
			fprintf(fout,"%0.2f	",qt->stats[i][j]);
		}
		fprintf(fout,"\n");
		//	}
		//}
	}
	if(outfile){
		fclose(fout);
	}
}

void report_stats(struct query_table* qt,struct parameters* param)
{
	int i;//,j;
	//FILE *fout = NULL;
		
	//char *names[] = { "Total","Exact","M","INDEL","MM","MMM"};
	fprintf(stderr,"\n	Mapping Report:\n");
	fprintf(stderr,"	%1.0f total (%d unique)\n",qt->total_expression,qt->numseq);
	
	if(param->upper_limit == 1){
		//fprintf(fout,"	Mapped	0	1	>1		Mapped	0	1	>1\n");
		i = 0;
		fprintf(stderr,"	Total mapped:	%0.2f%% (%0.2f)\n",(qt->stats[i][1]+qt->stats[i][2])/qt->total_expression*100,qt->stats[i][1]+qt->stats[i][2]);
		fprintf(stderr,"		1:	%0.2f%% (%0.2f)\n",(qt->stats[i][1])/qt->total_expression*100,qt->stats[i][1]);
		fprintf(stderr,"		>1:	%0.2f%% (%0.2f)\n",(qt->stats[i][2])/qt->total_expression*100,qt->stats[i][2]);
		if(qt->stats[i][2]){
			fprintf(stderr,"\n	Comment: %0.2f tags accounting for %0.2f%% of the expression are excluded from the mapping file because they map more than %d times ",qt->stats[i][2],(qt->stats[i][2])/qt->total_expression*100,param->upper_limit);
			if(param->upper_limit == 10){
				fprintf(stderr,"(the default for the -u parameter).\n\n");
			}else{
				fprintf(stderr,"(as specified via the -u parameter).\n\n");
			}
		}
		/*i = 0;
		
		fprintf(fout,"%s	",names[i]);
		fprintf(fout,"%0.2f%%	",(qt->stats[i][1]+qt->stats[i][2])/qt->total_expression*100);
		for(j = 0; j < 3;j++){
			fprintf(fout,"%0.2f%%	",qt->stats[i][j]/qt->total_expression*100);
		}
		
		fprintf(fout,"%s	",names[i]);
		fprintf(fout,"%0.2f	",qt->stats[i][1]+qt->stats[i][2]);
		for(j = 0; j < 3;j++){
			fprintf(fout,"%0.2f	",qt->stats[i][j]);
		}
		fprintf(fout,"\n");*/
		//	}
		//}
	}else{
		i = 0;
		fprintf(stderr,"	Total mapped:	%0.2f%% (%0.2f)\n",(qt->stats[i][1]+qt->stats[i][2]+qt->stats[i][3])/qt->total_expression*100,qt->stats[i][1]+qt->stats[i][2]+qt->stats[i][3]);
		fprintf(stderr,"		1:	%0.2f%% (%0.2f)\n",(qt->stats[i][1])/qt->total_expression*100,qt->stats[i][1]);
		fprintf(stderr,"		2-%d:	%0.2f%% (%0.2f)\n",param->upper_limit,(qt->stats[i][2])/qt->total_expression*100,qt->stats[i][2]);
		fprintf(stderr,"		>%d:	%0.2f%% (%0.2f)\n",param->upper_limit,(qt->stats[i][3])/qt->total_expression*100,qt->stats[i][3]);
		if(qt->stats[i][3]){	
			fprintf(stderr,"\n	Comment: %0.2f tags accounting for %0.2f%% of the expression are excluded from the mapping file because they map more than %d times ",qt->stats[i][3],(qt->stats[i][3])/qt->total_expression*100,param->upper_limit);
			if(param->upper_limit == 10){
				fprintf(stderr,"(the default for the -u parameter).\n\n");
			}else{
				fprintf(stderr,"(as specified via the -u parameter).\n\n");
			}
		}
	}
	fprintf(stderr,"\n");
}


void print_ann_score(struct query_table* qt,int max_no_ann)
{
	double* ab = 0;
	ab = malloc(sizeof(double)*2);
	get_ann_score(qt->tree,ab,max_no_ann);
	fprintf(stderr,"Score:%f\n",ab[0]/ab[1]*(ab[0]+ab[1])); 
	free(ab);
}

void print_ann(struct query_table* qt,struct parameters* param)
{

	FILE *fout = NULL;
	if(param->ann){
		if(!strcmp(param->ann,"STDOUT")){
			fout = stdout; 
		}else{
			if ((fout = fopen(param->ann, "w")) == NULL){
				fprintf(stderr,"can't open output\n");
				exit(-1);
			}
		}
	}else{
		fout = stdout; 
	}
	fprintf(fout,"Database(s)	Percentage	Counts\n");	
	if(param->types){
		print_tree(qt->tree,param->types,qt->total_expression,fout);
	}else{
		print_tree(qt->tree,param->infile,qt->total_expression,fout);
	}
	if(param->ann){
		fclose(fout);
	}
}

void print_unmapped_fasta(struct query_table* qt,char* outfile)
{
	int i;
	//char sn[MAX_LINE];
	static int print = 0;
	FILE *fout = NULL;
	if(outfile){
		if(!strcmp(outfile,"STDOUT")){
			fout = stdout; 
		}else{
			if(!print){
				if ((fout = fopen(outfile, "w")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}else{
				if ((fout = fopen(outfile, "a+")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}
		}
	}else{
		fout = stdout; 
	}
	for(i= 0; i < qt->size;i++){
		if(qt->park[i]->stats[0] == 0){
			if(!qt->park[i]->sn){
				fprintf(fout,">%s_%0.0f\n%s\n",qt->park[i]->seq,qt->park[i]->count,qt->park[i]->seq);
			}else{
				fprintf(fout,">%s\n%s\n",qt->park[i]->sn,qt->park[i]->seq);
			}
			//sprintf(sn, "%s_%f",qt->park[i]->seq,qt->park[i]->count);
			//fprintf(fout,">%s\n%s\n",sn,qt->park[i]->seq);
		}
	}
	if(outfile){
		fclose(fout);
	}
	print++;
}

void print_multi_mapped_fasta(struct query_table* qt,char* outfile,const unsigned int ulimit)
{
	int i;
	static int print = 0;
	//char sn[MAX_LINE];
	FILE *fout = NULL;
	if(outfile){
		if(!strcmp(outfile,"STDOUT")){
			fout = stdout; 
		}else{
			if(!print){
				if ((fout = fopen(outfile, "w")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}else{
				if ((fout = fopen(outfile, "a+")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}
		}
	}else{
		fout = stdout; 
	}
	for(i= 0; i < qt->size;i++){
		if(qt->park[i]->stats[0] > ulimit){
			if(!qt->park[i]->sn){
				fprintf(fout,">%s_%0.0f\n%s\n",qt->park[i]->seq,qt->park[i]->count,qt->park[i]->seq);
			}else{
				fprintf(fout,">%s\n%s\n",qt->park[i]->sn,qt->park[i]->seq);
			}
			//sprintf(sn, "%s_%f",qt->park[i]->seq,qt->park[i]->count);
			//fprintf(fout,">%s\n%s\n",sn,qt->park[i]->seq);
		}
	}
	if(outfile){
		fclose(fout);
	}
	print++;
}

void print_mapped_fasta(struct query_table* qt,char* outfile,unsigned int ulimit)
{
	int i;
	FILE *fout = NULL;
	static int print = 0;

	if (!ulimit){
		ulimit = UINT_MAX;
	}
	
	if(outfile){
		if(!strcmp(outfile,"STDOUT")){
			fout = stdout; 
		}else{
			if(!print){
				if ((fout = fopen(outfile, "w")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}else{
				if ((fout = fopen(outfile, "a+")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}
		}
	}else{
		fout = stdout; 
	}
	for(i= 0; i < qt->size;i++){
		if(qt->park[i]->stats[0] > 0 && qt->park[i]->stats[0] <= ulimit){
			if(!qt->park[i]->sn){
				fprintf(fout,">%s_%0.0f\n%s\n",qt->park[i]->seq,qt->park[i]->count,qt->park[i]->seq);
			}else{
				fprintf(fout,">%s\n%s\n",qt->park[i]->sn,qt->park[i]->seq);
			}
		}
	}
	if(outfile){
		fclose(fout);
	}
	print++;
}

void print_counts(struct query_table* qt,struct parameters* param,int db)
{
	char *tmp = 0;
	//char sn[MAX_LINE];
	//char* q_name;
	static int print = 0; 
	int i;
	int j;
	
	FILE *fout = NULL;
	if(param->count_file){
		if(!strcmp(param->count_file,"STDOUT")){
			fout = stdout; 
		}else{
			if(!print){
				if ((fout = fopen(param->count_file, "w")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}else{
				if ((fout = fopen(param->count_file, "a+")) == NULL){
					fprintf(stderr,"can't open output\n");
					exit(-1);
				}
			}
		}
	}else{
		fout = stdout; 
	}
	
	if(db == -1){
		fprintf(fout,"Database:");	
		for(i = 0;i < param->infiles-1;i++){
			tmp = shorten_name(param->infile[i]);
			fprintf(fout,"%s,",tmp);
		}
		tmp = shorten_name(param->infile[param->infiles-1]);
		fprintf(fout,"%s",tmp);
		fprintf(fout,"\n");
	}else{
		tmp = shorten_name(param->infile[db]);
		fprintf(fout,"Database:%s\n",tmp);
	}
	fprintf(fout,"Name	Total	Exact	M	INDEL	MM	MMM\n");
	
	for(i= 0; i < qt->size;i++){
		if(!qt->park[i]->sn){
			//sprintf(sn,"%s_%f",qt->park[i]->seq,qt->park[i]->count);
			//q_name = sn;
			fprintf(fout,"%s_%0.2f	",qt->park[i]->seq,qt->park[i]->count);
		}else{
			fprintf(fout,"%s	",qt->park[i]->sn);
			//q_name = qt->park[i]->sn;
		}
		//sprintf(sn, "%s_%f",qt->park[i]->seq,qt->park[i]->count);
		//fprintf(fout,"%s	",q_name);
		for(j =0;j < 6;j++){
			fprintf(fout,"%d	",qt->park[i]->stats[j]);
		}
		fprintf(fout,"\n");
	}
	if(param->count_file){
		fclose(fout);
	}
	print++;
}


void clean_name_mapping_output(struct query_table* qt,struct parameters* param)
{
	struct query* q = 0;
	FILE *file = 0;
	char line[MAX_LINE];
	char tmp_seq[MAX_LINE];
	int i;
	int len = 0;
	static int write = 0;
	char *time_s = 0;
	time_s = get_time(time_s);

	for(i= 0; i < qt->size;i++){
		if(qt->park[i]->stats[0] > 0 && qt->park[i]->stats[0] <= param->upper_limit){
			param->mapped_expression += qt->park[i]->count;
		}
	}
	
	
	if (!(file = fopen(param->outfile_tmp, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
	}
	if(write){
		if ((foutfile = fopen(param->outfile, "a")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
	}else{
		if ((foutfile = fopen(param->outfile, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
		//fprintf(foutfile,"##%s\n",OUTPUT_VERSION);
		//fprintf(foutfile,"##source-version %s\n",VERSION);
		//fprintf(foutfile,"##date %s\n",time_s);
	}
	
	while(fgets(line, MAX_LINE, file)){
		//fprintf(stderr,"ECHO %s",line);
		len = 0;
		for(i = 0;i < MAX_LINE;i++){
			if(line[i] == '\t'){
				tmp_seq[len] = 0;
				break;
			}
			tmp_seq[len] = line[i];
			len++;
		}
		//fprintf(stderr,"Searching with:%s\n",tmp_seq);
		q = lookup_name_hash(qt,tmp_seq,len);
		//q = search_hash(qt,q,tmp_seq);
		if(q){
			if(q->stats[0] <= param->upper_limit){
				//fprintf(foutfile,"%s",line);
				line[strlen(line)-1] = 0;
				fprintf(foutfile,"%s	%0.2f	%d\n",line,q->count,q->stats[0]);
				
				//test += q->count;
			//	fprintf(stderr,"Included:%s",line);
			}
		}else{
			fprintf(stderr,"Weird not FOUND:%s",line);
		}
		
	}
	//fprintf(stderr,"%f\n",test);
	write++;
	fclose(foutfile);
	fclose(file);
}

void clean_mapping_output(struct query_table* qt,struct parameters* param)
{
	struct query* q = 0;
	FILE *file = 0;
	char line[MAX_LINE];
	char tmp_seq[MAX_LINE];
	int i;
	int len = 0;
	static int write = 0;
	char *time_s = 0;
	time_s = get_time(time_s);
	//double test = 0.0;
	
	for(i= 0; i < qt->size;i++){
		if(qt->park[i]->stats[0] > 0 && qt->park[i]->stats[0]<= param->upper_limit){
			param->mapped_expression += qt->park[i]->count;
		}
	}
	
	if (!(file = fopen(param->outfile_tmp, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
	}
	if(write){
		if ((foutfile = fopen(param->outfile, "a")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
	}else{
		if ((foutfile = fopen(param->outfile, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
		//fprintf(foutfile,"##%s\n",OUTPUT_VERSION);
		//fprintf(foutfile,"##source-version %s\n",VERSION);
		//fprintf(foutfile,"##date %s\n",time_s);
	}
		
	while(fgets(line, MAX_LINE, file)){
		//fprintf(stderr,"ECHO %s",line);
		len = 0;
		for(i = 0;i < MAX_LINE;i++){
			if(line[i] == '\t' || line[i] == '_'){
				tmp_seq[len] = 0;
				break;
			}
			tmp_seq[len] = line[i];
			len++;
		}
		//fprintf(stderr,"Searching with:%s\n",tmp_seq);
		q = lookup_hash(qt,tmp_seq,len);
		if(q){
			if(q->stats[0] <= param->upper_limit){
			//	fprintf(stderr,"Included:%s",line);
				line[strlen(line)-1] = 0;
				fprintf(foutfile,"%s	%0.2f	%d\n",line,q->count,q->stats[0]);
				//param->mapped_expression += q->count;
				
			}
		}else{
			fprintf(stderr,"Weird not FOUND:%s",line);
		}
		
	}
	//fprintf(stderr,"%f\n",test);
	fclose(foutfile);
	fclose(file);
	write++;
}


void copy_mapping_output(struct query_table* qt,struct parameters* param)
{
	FILE *file = 0;
	char line[MAX_LINE];
	static int write = 0;
	char *time_s = 0;
	time_s = get_time(time_s);
	
	if (!(file = fopen(param->outfile_tmp, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
	}
	if(write){
		//fprintf(stderr,"Appending.....\n");
		if ((foutfile = fopen(param->outfile, "a")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
	}else{
		//fprintf(stderr,"NEW FILE ....\n");
		if ((foutfile = fopen(param->outfile, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
		//fprintf(foutfile,"##%s\n",OUTPUT_VERSION);
		//fprintf(foutfile,"##source-version %s\n",VERSION);
		//fprintf(foutfile,"##date %s\n",time_s);
	}
	
	while(fgets(line, MAX_LINE, file)){
		fprintf(foutfile,"%s",line);
	}
	fclose(foutfile);
	fclose(file);
	write++;
}



















