/*
	pattern_searching.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2006 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"
#include "pattern_searching.h"
#include "mapping_output.h"


#ifdef thread

void* pattern_matching_thread(void *threadarg)
{
	struct thread_data *data;
	data = (struct thread_data *) threadarg;

	const struct query_table* qt = data->q;
	const struct fasta* t = data->t;
	
	const unsigned int upper_limit = data->param->upper_limit;
	const int start = data->start;
	const int end = data->end;
	const int format = data->param->format;
	const int mode = data->param->mode;
	struct query** tmp = qt->park;
	
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	int i;
	sp->mis = 0;
	sp->indel = 0;
	
	
	if(mode & 1){
		for(i = start; i < end;i++){
			sp = pm(sp,t,tmp[i]->seq,tmp[i]->len);
			tmp[i]->stats[0] += sp->sum;
			tmp[i]->stats[1] += sp->sum;
			if(sp->sum > 0 && tmp[i]->stats[0] <= upper_limit){
				mapping_output(sp,t,tmp[i],"E",format);
			}
		}
	}
	//char rc[MAX_LINE];
	
	if(mode & 2){
		for(i = start; i < end;i++){
			pm_n_mis(t,tmp[i],tmp[i]->seq,upper_limit,format,0,data->param->mismatches,0,data->param->mismatches+1);
			//reverse_complement(tmp[i]->seq,rc,tmp[i]->len);
			//pm_n_mis(t,tmp[i],rc,upper_limit,format,0,data->param->mismatches,0,data->param->mismatches+1);
		}
	}
	if(mode & 4){
		for(i = start; i < end;i++){
			pm_ins(t,tmp[i],upper_limit,format);
			pm_del(t,tmp[i],upper_limit,format);
		}
	}
	
	free(sp);
	//return NULL;
	pthread_exit((void *) 0);
}

void* pattern_matching_quick_thread(void *threadarg)
{
	struct thread_data *data;
	data = (struct thread_data *) threadarg;

	const struct query_table* qt = data->q;
	const struct fasta* t = data->t;
	
	const unsigned int upper_limit = data->param->upper_limit;
	const int start = data->start;
	const int end = data->end;
	const int format = data->param->format;
	const int mode = data->param->mode;
	struct query** tmp = qt->park;
	
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	int i;
	sp->mis = 0;
	sp->indel = 0;
	
	
	if(mode & 1){
		for(i = start; i < end;i++){
			if(tmp[i]->stats[0] <= upper_limit){
				sp = pm(sp,t,tmp[i]->seq,tmp[i]->len);
				tmp[i]->stats[0] += sp->sum;
				//tmp[i]->stats[1] += sp->sum;
				if(sp->sum > 0 && tmp[i]->stats[0] <= upper_limit){
					mapping_output(sp,t,tmp[i],"E",format);
				}
			}
		}
	}
	//char rc[MAX_LINE];
	
	if(mode & 2){
		for(i = start; i < end;i++){
			if(tmp[i]->stats[0] <= upper_limit){
				pm_n_mis_quick(t,tmp[i],tmp[i]->seq,upper_limit,format,0,data->param->mismatches,0,data->param->mismatches+1);
			}
			//reverse_complement(tmp[i]->seq,rc,tmp[i]->len);
			//pm_n_mis(t,tmp[i],rc,upper_limit,format,0,data->param->mismatches,0,data->param->mismatches+1);
		}
	}
	if(mode & 4){
		for(i = start; i < end;i++){
			pm_ins(t,tmp[i],upper_limit,format);
			pm_del(t,tmp[i],upper_limit,format);
		}
	}
	
	free(sp);
	//return NULL;
	pthread_exit((void *) 0);
}


#else

void pattern_matching(struct query_table* qt,const struct fasta* t,const struct parameters* param)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp !=0);
	const unsigned int upper_limit = param->upper_limit;
	const int format = param->format;
	const int mode = param->mode;
	struct query** tmp = qt->park;

	int i,a;
	int len = 0;
	sp->mis = 0;
	sp->indel = 0;
	//a = 1;
	if(mode & 1){
		for(i = 0; i < qt->size;i++){
			len = tmp[i]->len;
			sp = pm(sp,t,tmp[i]->seq,len);
			tmp[i]->stats[0] += sp->sum;
			tmp[i]->stats[1] += sp->sum;
			if(sp->sum > 0 &&  tmp[i]->stats[0] <= upper_limit){
				mapping_output(sp,t,tmp[i],"E",format);
			}
			//fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
			//a++;
		}
		//fprintf(stderr,"\n");
	}
	
	//a = 1;
	if(mode & 2){
		for(i = 0; i < qt->size;i++){
			pm_n_mis(t,tmp[i],tmp[i]->seq,upper_limit,format,0,param->mismatches,0,param->mismatches+1);
		//	fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
		//	a++;
		}
		//fprintf(stderr,"\n");
	}
	
	a = 1;
	if(mode & 4){
		for(i = 0; i < qt->size;i++){

			pm_ins(t,tmp[i],upper_limit,format);
			pm_del(t,tmp[i],upper_limit,format);
			//fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
			a++;
		}
		//fprintf(stderr,"\n");
	}
	
	free(sp);
}


void pattern_matching_quick(struct query_table* qt,const struct fasta* t,const struct parameters* param)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp !=0);
	const unsigned int upper_limit = param->upper_limit;
	const int format = param->format;
	const int mode = param->mode;
	struct query** tmp = qt->park;

	int i,a;
	int len = 0;
	sp->mis = 0;
	sp->indel = 0;
	//a = 1;
	if(mode & 1){
		for(i = 0; i < qt->size;i++){
			if(tmp[i]->stats[0] <= upper_limit){
			len = tmp[i]->len;
			sp = pm(sp,t,tmp[i]->seq,len);
			tmp[i]->stats[0] += sp->sum;
			tmp[i]->stats[1] += sp->sum;
			if(sp->sum > 0 &&  tmp[i]->stats[0] <= upper_limit){
				mapping_output(sp,t,tmp[i],"E",format);
			}
	//		fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
	//		a++;
			}
		}
	//	fprintf(stderr,"\n");
	}
	
	//a = 1;
	if(mode & 2){
		for(i = 0; i < qt->size;i++){
			if(tmp[i]->stats[0] <= upper_limit){
			pm_n_mis_quick(t,tmp[i],tmp[i]->seq,upper_limit,format,0,param->mismatches,0,param->mismatches+1);
			}
	//		fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
	//		a++;
		}
	//	fprintf(stderr,"\n");
	}
	
	//a = 1;
	if(mode & 4){
		for(i = 0; i < qt->size;i++){

			pm_ins(t,tmp[i],upper_limit,format);
			pm_del(t,tmp[i],upper_limit,format);
	//		fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
	//		a++;
		}
	//	fprintf(stderr,"\n");
	}
	
	free(sp);
}

#endif


struct suffix_pos* pm(struct suffix_pos* sp,const struct fasta*t ,const char *p,const int len)
{
	const int rev[26] = {0,0,1,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,3,0,0,0,0,0,0};
	char rc[MAX_LINE];
	unsigned int i = 0;
	unsigned int r = 0;
	int j;

	reverse_complement2(p,rc,len);

	
	sp->e_p = 0;
	sp->s_p = 0;
	sp->e_m = 0;
	sp->s_m = 0;
	sp->sum = 0;
	for(j = 0; j < 16;j++){
		i = i << 2;
		r = r << 2;
		i |= rev[p[j]-65];
		r |= rev[rc[j]-65];
	}
	
	
	//i = tag_index2(p);
	if(t->mer_hash[i>>5] & (1<<(i & 0x1F))){
		sp->e_p = binsearch_up(p,t->string,t->suffix,len,t->string_len);		
		sp->s_p = binsearch_down(p,t->string,t->suffix,len,sp->e_p);
		sp->sum = sp->e_p-sp->s_p;
	}

	if(t->mer_hash[r>>5] & (1<<(r & 0x1F))){
		sp->e_m = binsearch_up(rc,t->string,t->suffix,len,t->string_len);
		sp->s_m = binsearch_down(rc,t->string,t->suffix,len,sp->e_m);
		sp->sum += (sp->e_m-sp->s_m);
	}
	return sp;
}

void pm_del(const struct fasta*t,struct query* q,const unsigned int upper_limit,const int format)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	char* tmp_seq = 0;
	char* mod_line = 0;
	int i,j,g;
	sp->mis = 0;
	sp->indel = 1;
	tmp_seq = malloc(sizeof(char)*(q->len));
	assert(tmp_seq != 0);
	mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	for(i = 0;i < q->len-1;i++){
		if(q->seq[i] != q->seq[i+1]){
			g = 0;// used to be 0
			for(j = 0;j < q->len;j++){//used to be 0
				if(j != i){
					tmp_seq[g] = q->seq[j];
					g++;
				}
			}
			tmp_seq[g] = 0;
			sp = pm(sp,t,tmp_seq,q->len-1);
			q->stats[0] += sp->sum;
			q->stats[3] += sp->sum;
			if(sp->sum > 0 && q->stats[0] <= upper_limit){
				sprintf(mod_line, "D:%d:%c",i,q->seq[i]);
				mapping_output(sp,t,q,mod_line,format);
			}
			
		}
	}
	g = 0;
	for(j = 0;j < q->len-1;j++){
		tmp_seq[g] = q->seq[j];
		g++;
	}
	tmp_seq[g] = 0;
	sp = pm(sp,t,tmp_seq,q->len-1);
	q->stats[0] += sp->sum;
	q->stats[3] += sp->sum;
	if(sp->sum > 0 && q->stats[0] <= upper_limit){
		sprintf(mod_line, "D:%d:%c",i,q->seq[i]);
		mapping_output(sp,t,q,mod_line,format);
	}
	
	
	free(tmp_seq);
	free(mod_line);
	free(sp);
}

void pm_ins(const struct fasta*t ,struct query* q,const unsigned int upper_limit,const int format)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp);
	char* alphabet = "ACGT";
	char* tmp_seq = 0;
	char* mod_line = 0;
	int i,j,f,g;
	sp->mis = 0;
	sp->indel = 1;
	tmp_seq = malloc(sizeof(char)*(q->len+2));
	assert(tmp_seq != 0);
	mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	for(i = 0;i < (q->len-2);i++){
		for(j = 0;j < 4;j++){
			if(q->seq[i+1] != alphabet[j]){
				g = 0;
				for(f = 0; f <= i;f++){
					tmp_seq[g] = q->seq[f];
					g++;
				}
				tmp_seq[g] = alphabet[j];
				g++;
				
				for(f = i+1;f < q->len;f++){
					tmp_seq[g] = q->seq[f];
					g++;
				}
				tmp_seq[g] = 0;
				sp = pm(sp,t,tmp_seq,q->len+1);
				q->stats[0] += sp->sum;
				q->stats[3] += sp->sum;
				if(sp->sum > 0 && q->stats[0] <= upper_limit){
					sprintf(mod_line, "I:%d:%c",i,alphabet[j]);
					mapping_output(sp,t,q,mod_line,format);
				}
			}
		}
	}
	i = q->len-2;
	for(j = 0;j < 4;j++){
		g = 0;
		for(f = 0; f <= i;f++){
			tmp_seq[g] = q->seq[f];
			g++;
		}
		tmp_seq[g] = alphabet[j];
		g++;
		
		for(f = i+1;f < q->len;f++){
			tmp_seq[g] = q->seq[f];
			g++;
		}
		tmp_seq[g] = 0;
		sp = pm(sp,t,tmp_seq,q->len+1);
		q->stats[0] += sp->sum;
		q->stats[3] += sp->sum;
		if(sp->sum > 0 && q->stats[0] <= upper_limit){
			sprintf(mod_line, "I%d:%c",i,alphabet[j]);
			mapping_output(sp,t,q,mod_line,format);
		}
		
	}
	free(tmp_seq);
	free(mod_line);
	free(sp);
}

void pm_n_mis(const struct fasta*t ,struct query* q,const char *p,const unsigned int upper_limit,const int format,int start,int n,char* mod_line, int mis)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	//assert(sp != 0);
	char* alphabet = "ACGT";
	char tmp_seq[MAX_LINE];
	int m_col[5] = {-1,-1,2,4,5};
	int i,j,c;//,f,g;
	
	int mod_len = 0;
	sp->mis = mis-n;
	sp->indel = 0;

	if(!mod_line){
		sp->indel = 1;
		mod_line = malloc(sizeof(char)*(10*n));
		assert(mod_line != 0);
	}else{
		mod_len = strlen(mod_line);
		mod_line[mod_len] = ',';
		mod_len++;
	}
	
	n--;
	c = m_col[mis-n];
	for(j =0; j < q->len;j++){
		tmp_seq[j] = p[j];
	}
	tmp_seq[j] = 0;
	
	/*int local_start = 0;
	if(q->len -start > 16){
		
		for(i = 0;i < q->len;i++){
			fprintf(stderr,"%c",tmp_seq[i]);
		}
		fprintf(stderr,"\n");
		
		fprintf(stderr,"prefix:");
		for(i = q->len-16;i < q->len;i++){
			fprintf(stderr,"%c",tmp_seq[i]);
		}
		fprintf(stderr,"\n");
		sp = pm(sp,t,tmp_seq+(q->len-16),16);
		if(!sp->sum){
			local_start = q->len-16;
		}
		
		
	}*/
	
	for(i = start;i < q->len;i++){
		if(i ==16){
			sp = pm(sp,t,tmp_seq,16);
			if(!sp->sum){
				break;
			}
		}
		for(j = 0;j < 4;j++){
			if(alphabet[j] != p[i]){
				//if(i >= local_start){
				tmp_seq[i] = alphabet[j];
				sp = pm(sp,t,tmp_seq,q->len);
				q->stats[c] += sp->sum;
				q->stats[0] += sp->sum;
				if(sp->sum > 0 && q->stats[0] <= upper_limit){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					mapping_output(sp,t,q,mod_line,format);
				}
				//}
				if(n){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					pm_n_mis(t,q,tmp_seq,upper_limit,format,i+1,n,mod_line,mis);
				}
			}
		}
		tmp_seq[i] = p[i];
	}
	if(sp->indel){
		free(mod_line);
	}
	free(sp);
}


void pm_n_mis_quick(const struct fasta*t ,struct query* q,const char *p,const unsigned int upper_limit,const int format,int start,int n,char* mod_line, int mis)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	//assert(sp != 0);
	char* alphabet = "ACGT";
	char tmp_seq[MAX_LINE];
	//int m_col[5] = {-1,-1,2,4,5};
	int i,j;//,c;//,f,g;
	
	int mod_len = 0;
	sp->mis = mis-n;
	sp->indel = 0;

	if(!mod_line){
		sp->indel = 1;
		mod_line = malloc(sizeof(char)*(10*n));
		assert(mod_line != 0);
	}else{
		mod_len = strlen(mod_line);
		mod_line[mod_len] = ',';
		mod_len++;
	}
	
	n--;
	//c = m_col[mis-n];
	for(j =0; j < q->len;j++){
		tmp_seq[j] = p[j];
	}
	tmp_seq[j] = 0;
	
	for(i = start;i < q->len;i++){
		if(i ==16){
			sp = pm(sp,t,tmp_seq,16);
			if(!sp->sum){
				break;
			}
		}
		
		for(j = 0;j < 4;j++){
			if(alphabet[j] != p[i]){
				tmp_seq[i] = alphabet[j];
				sp = pm(sp,t,tmp_seq,q->len);
				//q->stats[c] += sp->sum;				
				q->stats[0] += sp->sum;
				if(q->stats[0]  > upper_limit){
					goto endloop;
				}
				
				
				if(sp->sum > 0 && q->stats[0] <= upper_limit){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					mapping_output(sp,t,q,mod_line,format);
				}
				if(n){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					pm_n_mis(t,q,tmp_seq,upper_limit,format,i+1,n,mod_line,mis);
				}
			}
		}
		tmp_seq[i] = p[i];
	}
	
endloop:
	if(sp->indel){
		free(mod_line);
	}
	free(sp);
}
